Attribute VB_Name = "EventHandler_Global"
Option Explicit

'==================================================================
'
' Event Handler Control Support Module
'
' 1 October 2001
'
' You're free in using this software, as long as the
' orignal copyright/author information remains intact.
' In other words, respecting the person(s) who made it
' possible for you to view/learn something of this software.
'
' Copyright  2001 - Diederik van der Boor
' http://www.codingdomain.com
'
'==================================================================


' This module contains some global
' API declarations (for calling winprocs)
' and a winproc for the redirected events
' (AddressOf operator can only get)
' (proc addresses from a basic-module)

' Window Properties
Public Const WIN_PROP_CLSPTR As String = "EvtObjPtr"
Public Const WIN_PROP_OLDPROC As String = "EvtOldProc"
Public Declare Function GetProp Lib "user32" Alias "GetPropA" (ByVal hWnd As Long, ByVal lpString As String) As Long
Public Declare Function SetProp Lib "user32" Alias "SetPropA" (ByVal hWnd As Long, ByVal lpString As String, ByVal hData As Long) As Long
Public Declare Function RemoveProp Lib "user32" Alias "RemovePropA" (ByVal hWnd As Long, ByVal lpString As String) As Long

'' Window Is Local
'Private Declare Function GetWindowThreadProcessId Lib "user32" (ByVal hWnd As Long, lpdwProcessId As Long) As Long
'Private Declare Function GetCurrentProcessId Lib "kernel32" () As Long

' Local declarations when we need to unhook here (when class is destroyed/can't be accessed)
Private Const WM_DESTROY = &H2    ' Final message, kill all GDI Objects, some handles & most variables
Private Const WM_NCDESTROY = &H82 ' Form_Terminate?

' Set/Get the window proc handle
Public Const GWL_WNDPROC = (-4)
Public Declare Function IsWindow Lib "user32" (ByVal hWnd As Long) As Long
Public Declare Function GetWindowLong Lib "user32" Alias "GetWindowLongA" (ByVal hWnd As Long, ByVal nIndex As Long) As Long
Public Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" (ByVal hWnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Long) As Long

' Call Window Proc
Public Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Public Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcA" (ByVal lpPrevWndFunc As Long, ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Public Declare Function DefWindowProc Lib "user32" Alias "DefWindowProcA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long

' Pointers
Public Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (DestPtr As Any, SourcePtr As Any, ByVal NumOfBytes As Long)
Public Declare Function VarToPtr Lib "msvbvm50.dll" Alias "VarPtr" (Variable As Any) As Long
Public Declare Function IsBadReadPtr Lib "kernel32" (lp As Any, ByVal ucb As Long) As Long
Public Declare Function IsBadWritePtr Lib "kernel32" (lp As Any, ByVal ucb As Long) As Long
Public Declare Function IsBadCodePtr Lib "kernel32" (lpfn As Any) As Long
'Public Declare Function IsBadStringPtr Lib "kernel32" (lpsz As Any, ByVal ucchMax As Long) As Long

' Returns the address of a procedure passed by the AddressOf operator
Public Function ProcToPtr(ByVal AddressOf_Proc As Long) As Long
  ProcToPtr = AddressOf_Proc ' Why do we have to write nonesence like this? (Can't be just have an "Function(Pointer)" datatype?)
End Function

' Use object pointers to prevent VB from incrementing the class instance reference count
Public Function ObjFromPtr(ByVal ObjectPtr As Long) As Object
  Dim Obj As Object
  Dim ObjGet As Object                  ' Receiver for pointer
  
  ' These safety checks aren't 100% ok,
  ' but it will catch some really invalid pointers!
  If Not (CBool(IsBadCodePtr(ByVal ObjectPtr)) And CBool(IsBadReadPtr(ByVal ObjectPtr, 4)) And CBool(IsBadWritePtr(ByVal ObjectPtr, 4))) Then
    On Error Resume Next
      Call CopyMemory(ObjGet, ObjectPtr, 4) ' Turn the pointer into an ILLEGAL, UNCOUNTED interface
      Set Obj = ObjGet                      ' Assign to legal reference (Object references will be counted and registered by VB)
      Call CopyMemory(ObjGet, 0&, 4)        ' Destroy the illegal reference (Can't harm VB anymore)
    On Error GoTo 0
    Set ObjFromPtr = Obj                  ' Return legal object reference
  End If
End Function

' Window Proc of EventHandler class
Public Function EventHandler_WinProc(ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As EventResultMessage
  Dim ClsPtr As Long
  Dim PrevWinProc As Long
  Dim EvtHandler As EventHandler
  
  ' Get Window props
  PrevWinProc = GetProp(hWnd, WIN_PROP_OLDPROC)
  ClsPtr = GetProp(hWnd, WIN_PROP_CLSPTR)
  
  If PrevWinProc = 0 Then
    ' Couldn't do anything with this event information
    EventHandler_WinProc = evtIgnored
    Exit Function
  End If
  
  If (uMsg = WM_DESTROY Or uMsg = WM_NCDESTROY) Then ' And EventHandler_WndLocal(hWnd) Then
    ' Shit!?! (You Pressed the END Button/Used END Statement)
    ' Detach hook now or we risk a GPF! Calling some VB functions here will gpf u2!
    Call SetWindowLong(hWnd, GWL_WNDPROC, PrevWinProc)
    Call RemoveProp(hWnd, WIN_PROP_OLDPROC)
    Call RemoveProp(hWnd, WIN_PROP_CLSPTR)
    EventHandler_WinProc = CallWindowProc(PrevWinProc, hWnd, uMsg, wParam, lParam)
  Else
    If ClsPtr = 0 Then
      ' Always call the previous window proc
      EventHandler_WinProc = CallWindowProc(PrevWinProc, hWnd, uMsg, wParam, lParam)
    Else
      ' Try to contact the corresponding event handler class
      On Error Resume Next
      Set EvtHandler = ObjFromPtr(ClsPtr) ' Properly GPF/IPF when invalid object reference
      If EvtHandler Is Nothing Then
        On Error GoTo 0
        EventHandler_WinProc = CallWindowProc(PrevWinProc, hWnd, uMsg, wParam, lParam)
      Else
        On Error GoTo 0
        EventHandler_WinProc = EvtHandler.RaiseEventRM(hWnd, uMsg, wParam, lParam)
      End If
    End If
  End If
End Function

'Function EventHandler_WndLocal(ByVal hWnd As Long) As Boolean
'  Dim IDWnd As Long
'  Call GetWindowThreadProcessId(hWnd, IDWnd)
'  EventHandler_WndLocal = (IDWnd = GetCurrentProcessId())
'End Function

'Public Function ObjToPtr(Object As Object) As Long
'  Call CopyMemory(ObjToPtr, ByVal Object, 4) ' An object is just a long pointer (registered by VB as object type)
'End Function

